@@ -26,14 +26,12 @@
   (:args (object :scs (descriptor-reg))
          (value :scs (descriptor-reg any-reg null zero)))
   (:info name offset lowtag)
-  (:ignore name)
-  (:results)
+  (:temporary (:sc non-descriptor-reg) temp)
+  (:vop-var vop)
   (:generator 1
-    (storew value object offset lowtag)))
-
-(define-vop (init-slot set-slot)
-  (:info name dx-p offset lowtag)
-  (:ignore name dx-p))
+    (without-scheduling ()
+      (emit-gengc-barrier object nil temp (vop-nth-arg 1 vop) name)
+      (storew value object offset lowtag))))
 
 ;;;; Symbol hacking VOPs:
 
@@ -53,11 +51,10 @@
   (:temporary (:scs (non-descriptor-reg)) temp)
   (:temporary (:scs (descriptor-reg) :from (:argument 0)) obj-temp))
 
-;;; With Symbol-Value, we check that the value isn't the trap object.  So
-;;; Symbol-Value of NIL is NIL.
+;;; With Symbol-Value, we check that the value isn't the trap object.
 ;;;
 (define-vop (symbol-value checked-cell-ref)
-  (:translate symeval)
+  (:translate symbol-value)
   (:generator 9
     (move obj-temp object)
     (loadw value obj-temp symbol-value-slot other-pointer-lowtag)
@@ -67,19 +64,17 @@
       (inst nop))))
 
 ;;; Like CHECKED-CELL-REF, only we are a predicate to see if the cell is bound.
-(define-vop (boundp-frob)
+(define-vop (boundp)
   (:args (object :scs (descriptor-reg)))
   (:conditional)
   (:info target not-p)
   (:policy :fast-safe)
-  (:temporary (:scs (descriptor-reg)) value)
-  (:temporary (:scs (non-descriptor-reg)) temp))
-
-(define-vop (boundp boundp-frob)
+  (:temporary (:scs (non-descriptor-reg)) temp)
   (:translate boundp)
   (:generator 9
-    (loadw value object symbol-value-slot other-pointer-lowtag)
-    (inst xor temp value unbound-marker-widetag)
+    (inst lb temp object (+ (- (ash symbol-value-slot word-shift) other-pointer-lowtag)
+                            #+big-endian 3))
+    (inst xor temp temp unbound-marker-widetag)
     (if not-p
         (inst beq temp target)
         (inst bne temp target))
@@ -88,36 +83,17 @@
 (define-vop (fast-symbol-value cell-ref)
   (:variant symbol-value-slot other-pointer-lowtag)
   (:policy :fast)
-  (:translate symeval))
-
-(define-vop (symbol-hash)
-  (:policy :fast-safe)
-  (:translate symbol-hash)
-  (:args (symbol :scs (descriptor-reg)))
-  (:temporary (:scs (non-descriptor-reg)) temp)
-  (:results (res :scs (any-reg)))
-  (:result-types positive-fixnum)
-  (:generator 2
-    ;; The symbol-hash slot of NIL holds NIL because it is also the
-    ;; cdr slot, so we have to strip off the two low bits to make sure
-    ;; it is a fixnum.  The lowtag selection magic that is required to
-    ;; ensure this is explained in the comment in objdef.lisp
-    (loadw temp symbol symbol-hash-slot other-pointer-lowtag)
-    (inst srl temp n-fixnum-tag-bits)
-    (inst sll res temp n-fixnum-tag-bits)))
+  (:translate symbol-value))
 
 ;;; On unithreaded builds these are just copies of the non-global versions.
 (define-vop (%set-symbol-global-value set))
 (define-vop (symbol-global-value symbol-value)
-  (:translate sym-global-val))
+  (:translate symbol-global-value))
 (define-vop (fast-symbol-global-value fast-symbol-value)
-  (:translate sym-global-val))
+  (:translate symbol-global-value))
 
 ;;;; Fdefinition (fdefn) objects.
 
-(define-vop (fdefn-fun cell-ref)
-  (:variant fdefn-fun-slot other-pointer-lowtag))
-
 (define-vop (safe-fdefn-fun)
   (:translate safe-fdefn-fun)
   (:policy :fast-safe)
@@ -135,17 +111,16 @@
 
 (define-vop (set-fdefn-fun)
   (:policy :fast-safe)
-  (:translate (setf fdefn-fun))
-  (:args (function :scs (descriptor-reg) :target result)
+  (:args (function :scs (descriptor-reg))
          (fdefn :scs (descriptor-reg)))
   (:temporary (:scs (interior-reg)) lip)
   (:temporary (:scs (non-descriptor-reg)) type)
-  (:results (result :scs (descriptor-reg)))
   (:generator 38
+      (without-scheduling ()
+        (emit-gengc-barrier fdefn nil type)) ; type = temp
       (load-type type function (- fun-pointer-lowtag))
-      (inst nop)
       (inst xor type simple-fun-widetag)
-      (inst beq type normal-fn)
+      (inst beq type SIMPLE)
       (inst addu lip function
             (- (ash simple-fun-insts-offset word-shift)
                fun-pointer-lowtag))
@@ -159,23 +134,19 @@
       ;;   - lack of thread support makes a GC interrupt somewhat unlikely,
       ;;     (though any interrupt could call into lisp, also unlikely)
       (inst li lip (make-fixup 'closure-tramp :assembly-routine))
-    NORMAL-FN
+    SIMPLE
       (storew lip fdefn fdefn-raw-addr-slot other-pointer-lowtag)
-      (storew function fdefn fdefn-fun-slot other-pointer-lowtag)
-      (move result function)))
+      (storew function fdefn fdefn-fun-slot other-pointer-lowtag)))
 
 (define-vop (fdefn-makunbound)
   (:policy :fast-safe)
   (:translate fdefn-makunbound)
-  (:args (fdefn :scs (descriptor-reg) :target result))
+  (:args (fdefn :scs (descriptor-reg)))
   (:temporary (:scs (non-descriptor-reg)) temp)
-  (:results (result :scs (descriptor-reg)))
   (:generator 38
     (storew null-tn fdefn fdefn-fun-slot other-pointer-lowtag)
     (inst li temp (make-fixup 'undefined-tramp :assembly-routine))
-    (storew temp fdefn fdefn-raw-addr-slot other-pointer-lowtag)
-    (move result fdefn)))
-
+    (storew temp fdefn fdefn-raw-addr-slot other-pointer-lowtag)))
 
 
 ;;;; Binding and Unbinding.
@@ -189,20 +160,25 @@
   (:args (val :scs (any-reg descriptor-reg))
          (symbol :scs (descriptor-reg)))
   (:temporary (:scs (descriptor-reg)) temp)
+  (:temporary (:scs (non-descriptor-reg)) temp2)
   (:generator 5
     (loadw temp symbol symbol-value-slot other-pointer-lowtag)
     (inst addu bsp-tn bsp-tn (* 2 n-word-bytes))
     (storew temp bsp-tn (- binding-value-slot binding-size))
     (storew symbol bsp-tn (- binding-symbol-slot binding-size))
-    (storew val symbol symbol-value-slot other-pointer-lowtag)))
-
+    (without-scheduling ()
+      (emit-gengc-barrier symbol nil temp2)
+      (storew val symbol symbol-value-slot other-pointer-lowtag))))
 
 (define-vop (unbind)
   (:temporary (:scs (descriptor-reg)) symbol value)
+  (:temporary (:scs (non-descriptor-reg)) temp)
   (:generator 0
     (loadw symbol bsp-tn (- binding-symbol-slot binding-size))
     (loadw value bsp-tn (- binding-value-slot binding-size))
-    (storew value symbol symbol-value-slot other-pointer-lowtag)
+    (without-scheduling ()
+      (emit-gengc-barrier symbol nil temp)
+      (storew value symbol symbol-value-slot other-pointer-lowtag))
     (storew zero-tn bsp-tn (- binding-symbol-slot binding-size))
     (storew zero-tn bsp-tn (- binding-value-slot binding-size))
     (inst addu bsp-tn bsp-tn (* -2 n-word-bytes))))
@@ -212,28 +188,28 @@
   (:args (arg :scs (descriptor-reg any-reg) :target where))
   (:temporary (:scs (any-reg) :from (:argument 0)) where)
   (:temporary (:scs (descriptor-reg)) symbol value)
+  (:temporary (:scs (non-descriptor-reg)) temp)
   (:generator 0
-    (let ((loop (gen-label))
-          (skip (gen-label))
-          (done (gen-label)))
       (move where arg)
       (inst beq where bsp-tn done)
       (inst nop)
 
-      (emit-label loop)
+      LOOP
       (loadw symbol bsp-tn (- binding-symbol-slot binding-size))
       (inst beq symbol skip)
       (loadw value bsp-tn (- binding-value-slot binding-size))
-      (storew value symbol symbol-value-slot other-pointer-lowtag)
+      (without-scheduling ()
+        (emit-gengc-barrier symbol nil temp)
+        (storew value symbol symbol-value-slot other-pointer-lowtag))
       (storew zero-tn bsp-tn (- binding-symbol-slot binding-size))
 
-      (emit-label skip)
+      SKIP
       (storew zero-tn bsp-tn (- binding-value-slot binding-size))
       (inst addu bsp-tn bsp-tn (* -2 n-word-bytes))
       (inst bne where bsp-tn loop)
       (inst nop)
 
-      (emit-label done))))
+      DONE))
 
 
 
@@ -243,9 +219,9 @@
   closure-info-offset fun-pointer-lowtag
   (descriptor-reg any-reg) * %closure-index-ref)
 
-(define-full-setter set-funcallable-instance-info *
-  funcallable-instance-info-offset fun-pointer-lowtag
-  (descriptor-reg any-reg null zero) * %set-funcallable-instance-info)
+(define-full-setter %closure-index-set *
+  closure-info-offset fun-pointer-lowtag
+  (descriptor-reg any-reg null zero) * %closure-index-set)
 
 (define-full-reffer funcallable-instance-info *
   funcallable-instance-info-offset fun-pointer-lowtag
@@ -261,9 +237,13 @@
 (define-vop (closure-init)
   (:args (object :scs (descriptor-reg))
          (value :scs (descriptor-reg any-reg)))
-  (:info offset)
+  (:info offset dx)
+  (:temporary (:scs (non-descriptor-reg)) temp)
   (:generator 4
-    (storew value object (+ closure-info-offset offset) fun-pointer-lowtag)))
+    (without-scheduling ()
+      (unless dx
+        (emit-gengc-barrier object nil temp))
+      (storew value object (+ closure-info-offset offset) fun-pointer-lowtag))))
 
 (define-vop (closure-init-from-fp)
   (:args (object :scs (descriptor-reg)))
@@ -273,9 +253,6 @@
 
 ;;;; Value Cell hackery.
 
-(define-vop (value-cell-ref cell-ref)
-  (:variant value-cell-value-slot other-pointer-lowtag))
-
 (define-vop (value-cell-set cell-set)
   (:variant value-cell-value-slot other-pointer-lowtag))
 
@@ -306,16 +283,46 @@
 (define-full-reffer code-header-ref * 0 other-pointer-lowtag
   (descriptor-reg any-reg) * code-header-ref)
 
-(define-full-setter code-header-set * 0 other-pointer-lowtag
-  (descriptor-reg any-reg null zero) * code-header-set)
-
-
+(define-vop (code-header-set)
+  (:translate code-header-set)
+  (:policy :fast-safe)
+  (:args (object :scs (descriptor-reg))
+         (index :scs (any-reg))
+         (value :scs (any-reg descriptor-reg)))
+  (:arg-types * tagged-num *)
+  (:temporary (:scs (non-descriptor-reg)) temp card)
+  (:temporary (:sc non-descriptor-reg) pa-flag)
+  (:generator 10
+    (inst li temp (make-fixup "gc_card_table_mask" :foreign-dataref))
+    (loadw temp temp) ; address of gc_card_table_mask
+    (inst lw temp temp 0) ; value of gc_card_table_mask (4-byte int)
+    (pseudo-atomic (pa-flag)
+      ;; Compute card mark index
+      (inst srl card object gencgc-card-shift)
+      (inst and card card temp)
+      ;; Load mark table base
+      (inst li temp (make-fixup "gc_card_mark" :foreign-dataref)) ; address of linkage entry
+      (loadw temp temp) ; address of gc_card_mark
+      (loadw temp temp) ; value of gc_card_mark (pointer)
+      ;; Touch the card mark byte.
+      (inst add temp temp card)
+      (inst sb null-tn temp 0)
+      ;; set 'written' flag in the code header
+      ;; If two threads get here at the same time, they'll write the same byte.
+      (let ((byte (- #+little-endian 3 other-pointer-lowtag)))
+        (inst lbu temp object byte)
+        (inst or temp temp #x40)
+        (inst sb temp object byte))
+      ;; No need for LIP register because this is pseudo-atomic
+      (inst sll temp index (- word-shift n-fixnum-tag-bits))
+      (inst add temp object temp)
+      (inst sw value temp (- other-pointer-lowtag)))))
 
 ;;;; raw instance slot accessors
 
 (macrolet ((def (suffix sc primtype)
              `(progn
-                (define-vop (,(symbolicate "RAW-INSTANCE-REF/" suffix))
+                (define-vop ()
                   (:translate ,(symbolicate "%RAW-INSTANCE-REF/" suffix))
                   (:policy :fast-safe)
                   (:args (object :scs (descriptor-reg))
@@ -328,24 +335,21 @@
                     (inst addu lip object index)
                     (loadw value lip instance-slots-offset instance-pointer-lowtag)))
 
-                (define-vop (,(symbolicate "RAW-INSTANCE-SET/" suffix))
+                (define-vop ()
                   (:translate ,(symbolicate "%RAW-INSTANCE-SET/" suffix))
                   (:policy :fast-safe)
                   (:args (object :scs (descriptor-reg))
                          (index :scs (any-reg))
-                         (value :scs (,sc) :target result))
+                         (value :scs (,sc)))
                   (:arg-types * positive-fixnum ,primtype)
-                  (:results (result :scs (,sc)))
                   (:temporary (:scs (interior-reg)) lip)
-                  (:result-types ,primtype)
                   (:generator 5
                     (inst addu lip object index)
-                    (storew value lip instance-slots-offset instance-pointer-lowtag)
-                    (move result value))))))
+                    (storew value lip instance-slots-offset instance-pointer-lowtag))))))
   (def word unsigned-reg unsigned-num)
   (def signed-word signed-reg signed-num))
 
-(define-vop (raw-instance-ref/single)
+(define-vop ()
   (:translate %raw-instance-ref/single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
@@ -359,24 +363,20 @@
     (inst lwc1 value lip (- (* instance-slots-offset n-word-bytes)
                             instance-pointer-lowtag))))
 
-(define-vop (raw-instance-set/single)
+(define-vop ()
   (:translate %raw-instance-set/single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg))
-         (value :scs (single-reg) :target result))
+         (value :scs (single-reg)))
   (:arg-types * positive-fixnum single-float)
-  (:results (result :scs (single-reg)))
   (:temporary (:scs (interior-reg)) lip)
-  (:result-types single-float)
   (:generator 5
     (inst addu lip object index)
     (inst swc1 value lip (- (* instance-slots-offset n-word-bytes)
-                            instance-pointer-lowtag))
-    (unless (location= result value)
-      (inst fmove :single result value))))
+                            instance-pointer-lowtag))))
 
-(define-vop (raw-instance-ref/double)
+(define-vop ()
   (:translate %raw-instance-ref/double)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
@@ -398,16 +398,14 @@
         (:big-endian (inst lwc1 value lip immediate-offset))
         (:little-endian (inst lwc1-odd value lip immediate-offset))))))
 
-(define-vop (raw-instance-set/double)
+(define-vop ()
   (:translate %raw-instance-set/double)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg))
-         (value :scs (double-reg) :target result))
+         (value :scs (double-reg)))
   (:arg-types * positive-fixnum double-float)
-  (:results (result :scs (double-reg)))
   (:temporary (:scs (interior-reg)) lip)
-  (:result-types double-float)
   (:generator 5
     (inst addu lip object index)
     (let ((immediate-offset (- (* instance-slots-offset n-word-bytes)
@@ -419,11 +417,9 @@
                                instance-pointer-lowtag)))
       (ecase *backend-byte-order*
         (:big-endian (inst swc1 value lip immediate-offset))
-        (:little-endian (inst swc1-odd value lip immediate-offset))))
-    (unless (location= result value)
-      (inst fmove :double result value))))
+        (:little-endian (inst swc1-odd value lip immediate-offset))))))
 
-(define-vop (raw-instance-ref/complex-single)
+(define-vop ()
   (:translate %raw-instance-ref/complex-single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
@@ -445,36 +441,23 @@
           (- (* (1+ instance-slots-offset) n-word-bytes)
              instance-pointer-lowtag))))
 
-(define-vop (raw-instance-set/complex-single)
+(define-vop ()
   (:translate %raw-instance-set/complex-single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg))
-         (value :scs (complex-single-reg) :target result))
+         (value :scs (complex-single-reg)))
   (:arg-types * positive-fixnum complex-single-float)
-  (:results (result :scs (complex-single-reg)))
   (:temporary (:scs (interior-reg)) lip)
-  (:result-types complex-single-float)
   (:generator 5
     (inst addu lip object index)
-    (let ((value-real (complex-single-reg-real-tn value))
-          (result-real (complex-single-reg-real-tn result)))
-      (inst swc1
-            value-real
-            lip
-            (- (* instance-slots-offset n-word-bytes)
-               instance-pointer-lowtag))
-      (unless (location= result-real value-real)
-        (inst fmove :single result-real value-real)))
-    (let ((value-imag (complex-single-reg-imag-tn value))
-          (result-imag (complex-single-reg-imag-tn result)))
-      (inst swc1
-            value-imag
-            lip
-            (- (* (1+ instance-slots-offset) n-word-bytes)
-               instance-pointer-lowtag))
-      (unless (location= result-imag value-imag)
-        (inst fmove :single result-imag value-imag)))))
+    (inst swc1 (complex-single-reg-real-tn value)
+          lip (- (* instance-slots-offset n-word-bytes)
+                 instance-pointer-lowtag))
+    (inst swc1
+          (complex-single-reg-imag-tn value)
+          lip (- (* (1+ instance-slots-offset) n-word-bytes)
+                 instance-pointer-lowtag))))
 
 (define-vop (raw-instance-ref/complex-double)
   (:translate %raw-instance-ref/complex-double)
@@ -532,20 +515,17 @@
                               lip
                               immediate-offset))))))
 
-(define-vop (raw-instance-set/complex-double)
+(define-vop ()
   (:translate %raw-instance-set/complex-double)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg))
-         (value :scs (complex-double-reg) :target result))
+         (value :scs (complex-double-reg)))
   (:arg-types * positive-fixnum complex-double-float)
-  (:results (result :scs (complex-double-reg)))
   (:temporary (:scs (interior-reg)) lip)
-  (:result-types complex-double-float)
   (:generator 5
     (inst addu lip object index)
-    (let ((value-real (complex-double-reg-real-tn value))
-          (result-real (complex-double-reg-real-tn result)))
+    (let ((value-real (complex-double-reg-real-tn value)))
       (let ((immediate-offset (- (* instance-slots-offset n-word-bytes)
                                  instance-pointer-lowtag)))
         (ecase *backend-byte-order*
@@ -567,11 +547,8 @@
           (:little-endian (inst swc1-odd
                                 value-real
                                 lip
-                                immediate-offset))))
-      (unless (location= result-real value-real)
-        (inst fmove :double result-real value-real)))
-    (let ((value-imag (complex-double-reg-imag-tn value))
-          (result-imag (complex-double-reg-imag-tn result)))
+                                immediate-offset)))))
+    (let ((value-imag (complex-double-reg-imag-tn value)))
       (let ((immediate-offset (- (* (+ instance-slots-offset 2) n-word-bytes)
                                  instance-pointer-lowtag)))
         (ecase *backend-byte-order*
@@ -593,6 +570,4 @@
           (:little-endian (inst swc1-odd
                                 value-imag
                                 lip
-                                immediate-offset))))
-      (unless (location= result-imag value-imag)
-        (inst fmove :double result-imag value-imag)))))
+                                immediate-offset)))))))
