@@ -12,7 +12,7 @@
 #ifdef __linux__
 #define _GNU_SOURCE // for pthread_setname_np()
 #endif
-#include "sbcl.h"
+#include "genesis/sbcl.h"
 
 #include <stdlib.h>
 #include <stdio.h>
@@ -30,13 +30,14 @@
 #include "runtime.h"
 #include "validate.h"           /* for BINDING_STACK_SIZE etc */
 #include "thread.h"
+#include "genesis/thread.h"
 #include "arch.h"
 #include "target-arch-os.h"
 #include "os.h"
 #include "globals.h"
-#include "dynbind.h"
 #include "genesis/cons.h"
-#include "genesis/fdefn.h"
+#include "genesis/symbol.h"
+#include "genesis/instance.h"
 #include "genesis/vector.h"
 #include "interr.h"             /* for lose() */
 #include "gc.h"
@@ -106,6 +107,17 @@ unlink_thread(struct thread *th)
         th->next->prev = th->prev;
 }
 
+/* Not safe in general, but if your thread names are all
+ * simple-base-string and won't move, this is slightly ok */
+char* vm_thread_name(struct thread* th)
+{
+    if (!th) return "non-lisp";
+    struct thread_instance *lispthread = (void*)INSTANCE(th->lisp_thread);
+    lispobj name = lispthread->_name;
+    if (simple_base_string_p(name)) return vector_sap(name);
+    return "?";
+}
+
 #define get_thread_state(thread) \
  (int)__sync_val_compare_and_swap(&thread->state_word.state, -1, -1)
 
@@ -261,16 +273,6 @@ extern int arch_prctl(int code, unsigned long *addr);
 #define ASSOCIATE_OS_THREAD(thread) thread->os_thread = thread_self()
 #endif
 
-#ifndef LISP_FEATURE_SB_THREAD
-# define ASSIGN_CURRENT_THREAD(dummy)
-#elif defined LISP_FEATURE_GCC_TLS
-# define ASSIGN_CURRENT_THREAD(x) current_thread = x
-#elif !defined LISP_FEATURE_WIN32
-# define ASSIGN_CURRENT_THREAD(x) pthread_setspecific(current_thread, x)
-#else
-# define ASSIGN_CURRENT_THREAD(x) TlsSetValue(OUR_TLS_INDEX, x)
-#endif
-
 #ifdef LISP_FEATURE_WIN32
 // Need a function callable from assembly code, where the inline one won't do.
 void* read_current_thread() {
@@ -287,8 +289,8 @@ extern pthread_key_t ignore_stop_for_gc;
 #define COLLECT_GC_STATS
 #endif
 #ifdef COLLECT_GC_STATS
-static struct timespec gc_start_time;
-static long stw_elapsed,
+__attribute__((unused)) static struct timespec gc_start_time;
+__attribute__((unused)) static long stw_elapsed,
     stw_min_duration = LONG_MAX, stw_max_duration, stw_sum_duration,
     gc_min_duration = LONG_MAX, gc_max_duration, gc_sum_duration;
 int show_gc_stats, n_gcs_done;
@@ -392,6 +394,19 @@ void create_main_lisp_thread(lispobj function) {
     ASSIGN_CURRENT_THREAD(NULL);
 }
 
+void sb_posix_after_fork() { // for use by sb-posix:fork
+    struct thread* th = get_sb_vm_thread();
+    th->os_kernel_tid = get_nonzero_tid();
+#ifdef LISP_FEATURE_DARWIN
+    extern void darwin_reinit();
+    darwin_reinit();
+#endif
+#ifdef LISP_FEATURE_MARK_REGION_GC
+    extern void thread_pool_init();
+    thread_pool_init();
+#endif
+}
+
 #ifdef LISP_FEATURE_SB_THREAD
 
 void free_thread_struct(struct thread *th)
@@ -449,11 +464,27 @@ init_new_thread(struct thread *th,
 #endif
 }
 
+lispobj remset_transfer_list;
+
 static void
 unregister_thread(struct thread *th,
                   init_thread_data __attribute__((unused)) *scribble)
 {
     block_blockable_signals(0);
+#ifdef LISP_FEATURE_PERMGEN
+    lispobj my_remset = th->remset;
+    if (my_remset) {
+        lispobj tail = remset_transfer_list;
+        while (1) {
+            VECTOR(my_remset)->data[1] = tail;
+            lispobj actual_old = __sync_val_compare_and_swap(
+                &remset_transfer_list, tail, my_remset);
+            if (actual_old == tail) break;
+            tail = actual_old;
+        }
+        th->remset = 0;
+    }
+#endif
     gc_close_thread_regions(th, LOCK_PAGE_TABLE|CONSUME_REMAINDER);
 #ifdef LISP_FEATURE_SB_SAFEPOINT
     pop_gcing_safety(&scribble->safety);
@@ -608,7 +639,7 @@ void sb_set_os_thread_name(char* name)
      * but my testing showed that 12.1 seems to accept any length */
     pthread_set_name_np(pthread_self(), name);
 #endif
-#ifdef LISP_FEATURE_DARWIN
+#if defined LISP_FEATURE_DARWIN && !defined LISP_FEATURE_AVOID_PTHREAD_SETNAME_NP
     if (vector_len(v) < 64) pthread_setname_np(name);
 #endif
 }
@@ -776,7 +807,7 @@ static void detach_os_thread(init_thread_data *scribble)
         sigpending(&pending);
         if (sigismember(&pending, SIG_STOP_FOR_GC)) lose("clear stop-for-GC did not work");
 #else
-        int sig, rc;
+        __attribute__((unused)) int sig, rc;
         /* sigwait takes the mask of signals to allow through */
         rc = sigwait(&gc_sigset, &sig);
         gc_assert(rc == 0 && sig == SIG_STOP_FOR_GC);
@@ -930,8 +961,9 @@ alloc_thread_struct(void* spaces) {
 #ifdef THREAD_T_NIL_CONSTANTS_SLOT
     tls[THREAD_T_NIL_CONSTANTS_SLOT] = (NIL << 32) | LISP_T;
 #endif
-#ifdef THREAD_ALIEN_LINKAGE_TABLE_BASE_SLOT
-    tls[THREAD_ALIEN_LINKAGE_TABLE_BASE_SLOT] = (lispobj)ALIEN_LINKAGE_TABLE_SPACE_START;
+#ifdef LISP_FEATURE_LINKAGE_SPACE
+    tls[THREAD_LINKAGE_TABLE_SLOT] = (lispobj)linkage_space;
+    tls[THREAD_ALIEN_LINKAGE_TABLE_BASE_SLOT] = (lispobj)ALIEN_LINKAGE_SPACE_START;
 #endif
 #if defined LISP_FEATURE_X86_64 && defined LISP_FEATURE_LINUX
     tls[THREAD_MSAN_XOR_CONSTANT_SLOT] = 0x500000000000;
@@ -1009,15 +1041,11 @@ alloc_thread_struct(void* spaces) {
     th->state_word.user_thread_p = 1;
 
     lispobj* alien_stack_end = (lispobj*)((char*)th->alien_stack_start + ALIEN_STACK_SIZE);
-#ifdef ALIEN_STACK_GROWS_UPWARD
-    th->alien_stack_pointer = alien_stack_start;
-#elif defined LISP_FEATURE_X86 || defined LISP_FEATURE_X86_64
-    /* It is not necessary to subtract a word from the starting stack pointer
-     * because consuming C stack space pre-decrements before using the pointer.
-     * So this choice has nothing to do with saving a word of memory, and everything
-     * to do with my lack of understanding of the conventions for non-x86. */
+#if defined LISP_FEATURE_X86 || defined LISP_FEATURE_X86_64
+    // Alien-stack-pointer is predecremented upon use
     th->alien_stack_pointer = alien_stack_end;
 #else
+    // I do not know the convention for alien-stack-pointer
     th->alien_stack_pointer = alien_stack_end - 1;
 #endif
 
@@ -1070,8 +1098,6 @@ alloc_thread_struct(void* spaces) {
      * The thread alignment is BACKEND_PAGE_BYTES (from thread.h), but seeing as this is
      * a similar-but-different requirement, it pays to double-check */
     if ((lispobj)th & 0xFF) lose("Thread struct not at least 256-byte-aligned");
-    extern unsigned char* gc_card_mark;
-    th->card_table = (lispobj)gc_card_mark;
 #endif
 
 #ifdef LISP_FEATURE_SB_THREAD
@@ -1094,6 +1120,7 @@ alloc_thread_struct(void* spaces) {
     thread_extra_data(th)->synchronous_io_handle_and_flag = 0;
 #endif
     th->stepping = 0;
+    th->card_table = (lispobj)gc_card_mark;
     return th;
 }
 #ifdef LISP_FEATURE_SB_THREAD
@@ -1119,6 +1146,53 @@ uword_t create_thread(struct thread* th)
 int try_acquire_gc_lock() { return TryEnterCriticalSection(&in_gc_lock); }
 int release_gc_lock() { return mutex_release(&in_gc_lock); }
 
+static __attribute__((unused)) struct timespec stw_begin_realtime, stw_begin_cputime;
+long timespec_diff(struct timespec* begin, struct timespec* end)
+{
+#ifdef LISP_FEATURE_64_BIT
+    return (end->tv_sec - begin->tv_sec) * 1000000000L + (end->tv_nsec - begin->tv_nsec) ;
+#else
+    return (end->tv_sec - begin->tv_sec) * 1000000L + (end->tv_nsec - begin->tv_nsec) / 1000;
+#endif
+}
+#ifdef MEASURE_STOP_THE_WORLD_PAUSE
+void thread_accrue_stw_time(struct thread* th,
+                            struct timespec* begin_real,
+                            struct timespec* begin_cpu)
+{
+    /* A non-Lisp thread calling into Lisp via DEFINE-ALIEN-CALLABLE
+     * can receive SIG_STOP_FOR_GC as soon as it has a 'struct thread'
+     * and _before_ a thread instance has been consed */
+    if (th->lisp_thread) {
+        struct timespec now;
+        clock_gettime(CLOCK_MONOTONIC, &now);
+        unsigned long elapsed = timespec_diff(begin_real, &now);
+        struct thread_instance* ti = (void*)INSTANCE(th->lisp_thread);
+        if (elapsed > ti->uw_max_stw_pause) ti->uw_max_stw_pause = elapsed;
+        ti->uw_sum_stw_pause += elapsed;
+        ++ti->uw_ct_stw_pauses;
+        if (begin_cpu) {
+#ifdef CLOCK_THREAD_CPUTIME_ID
+          clock_gettime(CLOCK_THREAD_CPUTIME_ID, &now);
+          ti->uw_gc_virtual_time += timespec_diff(begin_cpu, &now);
+#endif
+        }
+    }
+}
+#endif
+
+#ifndef LISP_FEATURE_WIN32
+/* pthread_kill is not guaranteed to be reentrant, prevent
+ * gc_stop_the_world from interrupting another pthread_kill */
+int sb_thread_kill (pthread_t thread, int sig) {
+    sigset_t old;
+    block_blockable_signals(&old);
+    int ret = pthread_kill(thread, sig);
+    thread_sigmask(SIG_SETMASK, &old, NULL);
+    return ret;
+}
+#endif
+
 /* stopping the world is a two-stage process.  From this thread we signal
  * all the others with SIG_STOP_FOR_GC.  The handler for this signal does
  * the usual pseudo-atomic checks (we don't want to stop a thread while
@@ -1152,13 +1226,17 @@ int release_gc_lock() { return mutex_release(&in_gc_lock); }
  * (so that dereferencing was valid), but if dereferencing was valid, then the thread
  * can't have died (i.e. if ESRCH could be returned, then that implies that
  * the memory shouldn't be there) */
+
 void gc_stop_the_world()
 {
-#ifdef COLLECT_GC_STATS
-    struct timespec stw_begin_time, stw_end_time;
-    // Measuring the wait time has to use a realtime clock, not a thread clock
-    // because sleeping below in a sem_wait needs to accrue time.
-    clock_gettime(CLOCK_MONOTONIC, &stw_begin_time);
+#ifdef MEASURE_STOP_THE_WORLD_PAUSE
+    /* The thread performing stop-the-world does not use sig_stop_for_gc_handler on itself,
+     * so it would not accrue time spent stopped. Force it to, by considering it "paused"
+     * from the moment it wants to stop all other threads. */
+    clock_gettime(CLOCK_MONOTONIC, &stw_begin_realtime);
+#endif
+#ifdef CLOCK_THREAD_CPUTIME_ID
+    clock_gettime(CLOCK_THREAD_CPUTIME_ID, &stw_begin_cputime);
 #endif
     struct thread *th, *me = get_sb_vm_thread();
     int rc;
@@ -1193,12 +1271,6 @@ void gc_stop_the_world()
         }
     }
     event0("/gc_stop_the_world:end");
-#ifdef COLLECT_GC_STATS
-    clock_gettime(CLOCK_MONOTONIC, &stw_end_time);
-    stw_elapsed = (stw_end_time.tv_sec - stw_begin_time.tv_sec)*1000000000L
-                + (stw_end_time.tv_nsec - stw_begin_time.tv_nsec);
-    gc_start_time = stw_end_time;
-#endif
 }
 
 void gc_start_the_world()
@@ -1243,10 +1315,13 @@ void gc_start_the_world()
 
     lock_ret = mutex_release(&all_threads_lock);
     gc_assert(lock_ret);
+#ifdef MEASURE_STOP_THE_WORLD_PAUSE
+    thread_accrue_stw_time(me, &stw_begin_realtime, &stw_begin_cputime);
+#endif
 }
 
 #endif /* !LISP_FEATURE_SB_SAFEPOINT */
-#else
+#elif !defined STANDALONE_LDB
 // no threads
 void gc_stop_the_world() {}
 void gc_start_the_world() {}
@@ -1322,3 +1397,20 @@ void wake_thread(struct thread_instance* lispthread)
 #endif
 }
 #endif
+
+#ifdef LISP_FEATURE_ULTRAFUTEX
+extern int futex_wake(int *lock_word, int n);
+void lispmutex_wake_waiter()
+{
+    struct lispmutex* m = (void*)INSTANCE(read_TLS(CURRENT_MUTEX, get_sb_vm_thread()));
+    // The lock word is in the least-significant half of the state word if 64-bit.
+    // See the definition of MUTEX-STATE-ADDRESS which adds 4 if #+big-endian.
+    int* word =
+#ifdef LISP_FEATURE_BIG_ENDIAN
+                     1 +
+#endif
+        (int*)&m->uw_state;
+    *word = 0; // slam 0 in, meaning uncontested
+    futex_wake(word, 1);
+}
+#endif
